/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Lex configuration parser.
 *
 * This file defines the token for parsing the configuration file of waaad.
 * Note that each extension has a separate independant configuration file.
 *
 * Note : This module is NOT thread-safe. All processing must be done from one thread only.
 */

%{
/* Include the daemon's header files */
#include "waaad-internal.h"
/* Include yacc tokens definitions */
#include "conf-gram.h"

/* Update the column information */
#ifdef DEBUG_LEX
#define YY_USER_ACTION { 					\
	printf("(%d:%d-%d:%d)matched rule %d, lenght=%d, txt=%s\n", yylloc->first_line, yylloc->first_column, yylloc->last_line, yylloc->last_column, yy_act, yyleng, yytext); fflush(stdout);				\
	yylloc->first_column = yylloc->last_column + 1; 	\
	yylloc->last_column = yylloc->first_column + yyleng - 1;\
}
#else /* DEBUG_LEX */
#define YY_USER_ACTION { 					\
	yylloc->first_column = yylloc->last_column + 1; 	\
	yylloc->last_column = yylloc->first_column + yyleng - 1;\
}
#endif
%}

%option bison-bridge bison-locations
%option noyywrap

/* Use the following start condition to parse an URI */
%x	EXPECT_URI
%x	IN_URI

/* Quoted string. Multilines do not match. */
qstring		\"[^\"\n]*\"

%%

<*>\n			{ 
				/* Update the line count */
				yylloc->first_line++; 
				yylloc->last_line++; 
				yylloc->last_column=0; 
			} 

<*>[[:space:]]		; /* Eat all spaces */
<*>#.*$			; /* Eat all comments */

{qstring}		{
				/* First copy the string without the quotes for use in the yacc parser */
				yylval->string = strdup(yytext+1); /* This allocates one useless tail char but... it's easier :D */ 
				if (!yylval->string) {
					/* No more memory... */
					log_error("Unable to copy string '%s': %s\n", yytext, strerror(errno));
					return LEX_ERROR; /* trig an error in yacc parser */
				}
				yylval->string[yyleng-2] = '\0';
				
				/* the yacc parser will check the string is valid */
				return QSTRING;
			}
			
[[:digit:]]+		{
				/* Convert this to an integer value */
				int ret=0;
				ret = sscanf(yytext, "%i", &yylval->integer);
				if (ret != 1) {
					/* No matching: an error occurred */
					log_error("Unable to convert the value '%s' to a valid number: %s\n", yytext, strerror(errno));
					return LEX_ERROR; /* trig an error in yacc parser */
					/* Maybe we could REJECT instead of failing here? */
				}
				return INTEGER;
			}
				

	/* The GENERAL section tokens */
(?i:"[general]")	{ return SECTION_GENERAL; 			} /* enter the General section */

(?i:"Log")		{ return GEN_LOG; 				} /* How to specify the logging facility */
(?i:"console")		{ return GEN_LOG_CONSOLE; 			} /* Log messages to console */
(?i:"syslog")		{ return GEN_LOG_SYSLOG; 			} /* Log messages to syslog */

(?i:"Diameter-Id")	{ return GEN_DIAMID;				} /* How to specify the Diameter Identity of the local peer */

(?i:"LocalPort")	{ return GEN_PORT;				} /* The port the local peer is listening on */

(?i:"TcTimer")		{ return GEN_TCTIMER;				} /* Specify the Tc timer value */

(?i:"DisableRelay")	{ return GEN_DISABLERELAY;			} /* Is relaying disabled? */

	/* The EXTENSION section tokens */
(?i:"[extensions]")	{ return SECTION_EXTENSION; 			} /* enter the Extension section */

(?i:"filename")		{ return EXT_FILENAME;				}
(?i:"conffile")		{ return EXT_CONFFILE; 				}

	/* The PEERS section tokens */
(?i:"[peers]")		{ return SECTION_PEER;				} /* enter the Peers section */

(?i:"uri")		{ BEGIN(EXPECT_URI); return PEER_URI;		} /* We are now expecting an URI, we do not consider QSTRINGS anymore. */
<EXPECT_URI>"\""	{ BEGIN(IN_URI);     return '"';		} /* Enter the string describing the URI */
<IN_URI>"\""		{ BEGIN(INITIAL);    return '"';		} /* Exit the string describing the URI */
<IN_URI>"aaa://"	{ return PEER_SCHEME_NOSEC;			}
<IN_URI>"aaas://"	{ return PEER_SCHEME_SEC;			}
<IN_URI>[[][[:xdigit:]]*:[[:xdigit:]]*:[[:xdigit:].:]*[]]	{
				/* This matches a protected IPv6 or mapped ipv4 between [ ]. In this case, we remove the [ ] */
				yylval->string = strdup(yytext+1); /* This allocates one useless tail char but... it's easier :D */ 
				if (!yylval->string) {
					/* No more memory... */
					log_error("Unable to copy string '%s': %s\n", yytext, strerror(errno));
					return LEX_ERROR; /* trig an error in yacc parser */
				}
				yylval->string[yyleng-2] = '\0';
				
				/* the yacc parser will check the string is valid */
				return PEER_LOCATOR;
			}
<IN_URI>":"[[:digit:]]+	{ 
				/* Check this is a valid port number */
				int ret=0;
				ret = sscanf(yytext, ":%hu", &yylval->u16);
				if (ret != 1) {
					/* No matching: an error occurred */
					log_error("Unable to convert the value '%s' to a valid port number: %s\n", yytext, strerror(errno));
					return LEX_ERROR; /* trig an error in yacc parser */
				}
				return PEER_PORT;
			}
<IN_URI>(?i:";transport=sctp") 	{ return PEER_TRANS_SCTP;		}
<IN_URI>(?i:";transport=tcp") 	{ return PEER_TRANS_TCP;		}
<IN_URI>[[:alnum:].-]+	{
				/* Any string that did not match one of the previous IN_URI rules is treated as a locator */
				yylval->string = strdup(yytext);
				if (!yylval->string) {
					/* No more memory... */
					log_error("Unable to copy string '%s': %s\n", yytext, strerror(errno));
					return LEX_ERROR; /* trig an error in yacc parser */
				}
				
				/* the yacc parser will check the string is valid */
				return PEER_LOCATOR;
			}
	
	/* Valid single characters for yyparse */
<*>[=,;{}]		{ return yytext[0]; }

	/* Unrecognized token */
<*>[[:alnum:]]+		|	/* This rule is only useful to print a complete token in error messages */
	/* Unrecognized character */
<*>.			{ 
				log_error("Unrecognized text on line %d col %d: '%s'.\n", yylloc->first_line, yylloc->first_column, yytext);
			 	return LEX_ERROR; 
			}

%%
